Package: crisprDesign
Title: Comprehensive design of CRISPR gRNAs for nucleases and base
        editors
Version: 1.14.0
Authors@R: c(
    person("Jean-Philippe", "Fortin", email = "fortin946@gmail.com", role = c("aut", "cre")),
    person("Luke", "Hoberecht", email = "lukehob3@gmail.com", role = c("aut"))
    )
Description: Provides a comprehensive suite of functions to design and
        annotate CRISPR guide RNA (gRNAs) sequences. This includes on-
        and off-target search, on-target efficiency scoring, off-target
        scoring, full gene and TSS contextual annotations, and SNP
        annotation (human only). It currently support five types of
        CRISPR modalities (modes of perturbations): CRISPR knockout,
        CRISPR activation, CRISPR inhibition, CRISPR base editing, and
        CRISPR knockdown. All types of CRISPR nucleases are supported,
        including DNA- and RNA-target nucleases such as Cas9, Cas12a,
        and Cas13d. All types of base editors are also supported. gRNA
        design can be performed on reference genomes, transcriptomes,
        and custom DNA and RNA sequences. Both unpaired and paired gRNA
        designs are enabled.
Depends: R (>= 4.2.0), crisprBase (>= 1.1.3)
Imports: AnnotationDbi, BiocGenerics, Biostrings (>= 2.77.2), BSgenome
        (>= 1.77.1), crisprBowtie (>= 0.99.8), crisprScore (>= 1.15.2),
        GenomeInfoDb (>= 1.45.7), GenomicFeatures (>= 1.61.4),
        GenomicRanges (>= 1.61.1), IRanges, Matrix, MatrixGenerics,
        methods, reticulate, rtracklayer (>= 1.69.1), S4Vectors,
        Seqinfo, stats, txdbmaker (>= 1.5.6), utils, VariantAnnotation
        (>= 1.55.1)
Suggests: biomaRt, BSgenome.Hsapiens.UCSC.hg38,
        BSgenome.Mmusculus.UCSC.mm10, BiocStyle, crisprBwa (>= 0.99.7),
        knitr, rmarkdown, Rbowtie, Rbwa, RCurl, testthat
biocViews: CRISPR, FunctionalGenomics, GeneTarget
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
VignetteBuilder: knitr
BugReports: https://github.com/crisprVerse/crisprDesign/issues
URL: https://github.com/crisprVerse/crisprDesign
LazyData: true
Config/pak/sysreqs: make libbz2-dev libicu-dev liblzma-dev libpng-dev
        libxml2-dev libssl-dev python3 libx11-dev xz-utils zlib1g-dev
Repository: https://bioc-release.r-universe.dev
Date/Publication: 2026-04-28 12:58:44 UTC
RemoteUrl: https://github.com/bioc/crisprDesign
RemoteRef: RELEASE_3_23
RemoteSha: 78be86f80e1eb45b951e62947761bda406f17ca3
NeedsCompilation: no
Packaged: 2026-04-29 17:10:27 UTC; root
Author: Jean-Philippe Fortin [aut, cre],
  Luke Hoberecht [aut]
Maintainer: Jean-Philippe Fortin <fortin946@gmail.com>
Built: R 4.6.0; ; 2026-04-29 17:14:42 UTC; unix
