# Generated by roxygen2: do not edit by hand

export("alignments<-")
export("enzymeAnnotation<-")
export("geneAnnotation<-")
export("snps<-")
export("tssAnnotation<-")
export(GuideSet)
export(GuideSet2DataFrames)
export(PairedGuideSet)
export(TxDb2GRangesList)
export(addCompositeScores)
export(addConservationScores)
export(addCutSites)
export(addDistanceToTss)
export(addEditedAlleles)
export(addEditingSites)
export(addExonTable)
export(addGeneAnnotation)
export(addIsoformAnnotation)
export(addNtcs)
export(addOffTargetScores)
export(addOnTargetScores)
export(addOpsBarcodes)
export(addPamScores)
export(addPfamDomains)
export(addReinitiationFlag)
export(addRepeats)
export(addRestrictionEnzymes)
export(addSNPAnnotation)
export(addSequenceFeatures)
export(addSpacerAlignments)
export(addSpacerAlignmentsIterative)
export(addTssAnnotation)
export(addTxTable)
export(alignments)
export(bsgenome)
export(convertToMinMaxGRanges)
export(convertToProtospacerGRanges)
export(crisprNuclease)
export(customSequences)
export(cutLength)
export(designCompleteAnnotation)
export(designOpsLibrary)
export(editedAlleles)
export(enzymeAnnotation)
export(exonTable)
export(findSpacerPairs)
export(findSpacers)
export(flattenGuideSet)
export(geneAnnotation)
export(getBarcodeDistanceMatrix)
export(getConsensusIsoform)
export(getMrnaSequences)
export(getPAMSequence)
export(getPAMSiteFromStartAndEnd)
export(getPreMrnaSequences)
export(getSpacerAlignments)
export(getSpacerSequence)
export(getTssObjectFromTxObject)
export(getTxDb)
export(getTxInfoDataFrame)
export(mcols)
export(offTargets)
export(onTargets)
export(pamDistance)
export(pamOrientation)
export(pamSites)
export(preparePfamTable)
export(protospacers)
export(queryTss)
export(queryTxObject)
export(rankSpacers)
export(removeRepeats)
export(removeSpacersWithSecondaryTargets)
export(seqnames)
export(snps)
export(spacerDistance)
export(spacers)
export(targetOrigin)
export(tssAnnotation)
export(txTable)
export(updateOpsLibrary)
export(validateOpsLibrary)
exportClasses(GuideSet)
exportClasses(PairedGuideSet)
exportMethods("alignments<-")
exportMethods("enzymeAnnotation<-")
exportMethods("geneAnnotation<-")
exportMethods("snps<-")
exportMethods("tssAnnotation<-")
exportMethods(addCompositeScores)
exportMethods(addConservationScores)
exportMethods(addCutSites)
exportMethods(addDistanceToTss)
exportMethods(addEditingSites)
exportMethods(addGeneAnnotation)
exportMethods(addIsoformAnnotation)
exportMethods(addNtcs)
exportMethods(addOffTargetScores)
exportMethods(addOnTargetScores)
exportMethods(addPamScores)
exportMethods(addPfamDomains)
exportMethods(addRepeats)
exportMethods(addRestrictionEnzymes)
exportMethods(addSNPAnnotation)
exportMethods(addSequenceFeatures)
exportMethods(addSpacerAlignments)
exportMethods(addSpacerAlignmentsIterative)
exportMethods(addTssAnnotation)
exportMethods(alignments)
exportMethods(bsgenome)
exportMethods(crisprNuclease)
exportMethods(customSequences)
exportMethods(cutLength)
exportMethods(cutSites)
exportMethods(editedAlleles)
exportMethods(enzymeAnnotation)
exportMethods(exonTable)
exportMethods(geneAnnotation)
exportMethods(offTargets)
exportMethods(onTargets)
exportMethods(pamDistance)
exportMethods(pamLength)
exportMethods(pamOrientation)
exportMethods(pamSide)
exportMethods(pamSites)
exportMethods(pams)
exportMethods(protospacers)
exportMethods(prototypeSequence)
exportMethods(removeRepeats)
exportMethods(snps)
exportMethods(spacerDistance)
exportMethods(spacerLength)
exportMethods(spacers)
exportMethods(targetOrigin)
exportMethods(tssAnnotation)
exportMethods(txTable)
importClassesFrom(Biostrings,DNAStringSet)
importClassesFrom(GenomicRanges,GPos)
importClassesFrom(GenomicRanges,GRanges)
importClassesFrom(GenomicRanges,GRangesList)
importClassesFrom(IRanges,DataFrameList)
importClassesFrom(IRanges,IRanges)
importClassesFrom(S4Vectors,DataFrame)
importFrom(AnnotationDbi,select)
importFrom(BSgenome,getBSgenome)
importFrom(BSgenome,getSeq)
importFrom(BiocGenerics,"end<-")
importFrom(BiocGenerics,"rownames<-")
importFrom(BiocGenerics,"start<-")
importFrom(BiocGenerics,"strand<-")
importFrom(BiocGenerics,as.data.frame)
importFrom(BiocGenerics,cbind)
importFrom(BiocGenerics,colnames)
importFrom(BiocGenerics,end)
importFrom(BiocGenerics,invertStrand)
importFrom(BiocGenerics,organism)
importFrom(BiocGenerics,rbind)
importFrom(BiocGenerics,rownames)
importFrom(BiocGenerics,sort)
importFrom(BiocGenerics,start)
importFrom(BiocGenerics,strand)
importFrom(BiocGenerics,unlist)
importFrom(BiocGenerics,width)
importFrom(Biostrings,DNAString)
importFrom(Biostrings,DNAStringSet)
importFrom(Biostrings,DNA_ALPHABET)
importFrom(Biostrings,DNA_BASES)
importFrom(Biostrings,IUPAC_CODE_MAP)
importFrom(Biostrings,RNAStringSet)
importFrom(Biostrings,RNA_BASES)
importFrom(Biostrings,complement)
importFrom(Biostrings,extractAt)
importFrom(Biostrings,findPalindromes)
importFrom(Biostrings,getSeq)
importFrom(Biostrings,letterFrequency)
importFrom(Biostrings,matchPDict)
importFrom(Biostrings,matchPattern)
importFrom(Biostrings,palindromeLeftArm)
importFrom(Biostrings,reverse)
importFrom(Biostrings,reverseComplement)
importFrom(Biostrings,subseq)
importFrom(Biostrings,translate)
importFrom(GenomeInfoDb,"seqlevelsStyle<-")
importFrom(GenomeInfoDb,dropSeqlevels)
importFrom(GenomeInfoDb,keepStandardChromosomes)
importFrom(GenomeInfoDb,renameSeqlevels)
importFrom(GenomeInfoDb,seqlevelsStyle)
importFrom(GenomeInfoDb,seqnames)
importFrom(GenomicFeatures,as.list)
importFrom(GenomicFeatures,fiveUTRsByTranscript)
importFrom(GenomicFeatures,threeUTRsByTranscript)
importFrom(GenomicFeatures,tidyExons)
importFrom(GenomicFeatures,tidyIntrons)
importFrom(GenomicFeatures,tidyTranscripts)
importFrom(GenomicFeatures,transcriptLengths)
importFrom(GenomicRanges,GPos)
importFrom(GenomicRanges,GRanges)
importFrom(GenomicRanges,GRangesList)
importFrom(GenomicRanges,distanceToNearest)
importFrom(GenomicRanges,findOverlaps)
importFrom(GenomicRanges,flank)
importFrom(GenomicRanges,pintersect)
importFrom(GenomicRanges,pos)
importFrom(GenomicRanges,promoters)
importFrom(GenomicRanges,reduce)
importFrom(GenomicRanges,resize)
importFrom(GenomicRanges,trim)
importFrom(IRanges,"ranges<-")
importFrom(IRanges,IRanges)
importFrom(IRanges,findOverlaps)
importFrom(IRanges,pos)
importFrom(IRanges,shift)
importFrom(IRanges,trim)
importFrom(Matrix,Matrix)
importFrom(Matrix,rowSums)
importFrom(MatrixGenerics,rowRanges)
importFrom(S4Vectors,"first<-")
importFrom(S4Vectors,"mcols<-")
importFrom(S4Vectors,"metadata<-")
importFrom(S4Vectors,"second<-")
importFrom(S4Vectors,DataFrame)
importFrom(S4Vectors,List)
importFrom(S4Vectors,Pairs)
importFrom(S4Vectors,bindROWS)
importFrom(S4Vectors,first)
importFrom(S4Vectors,isTRUEorFALSE)
importFrom(S4Vectors,mcols)
importFrom(S4Vectors,metadata)
importFrom(S4Vectors,nchar)
importFrom(S4Vectors,queryHits)
importFrom(S4Vectors,second)
importFrom(S4Vectors,split)
importFrom(S4Vectors,subjectHits)
importFrom(Seqinfo,"genome<-")
importFrom(Seqinfo,"seqinfo<-")
importFrom(Seqinfo,"seqlevels<-")
importFrom(Seqinfo,Seqinfo)
importFrom(Seqinfo,checkCompatibleSeqinfo)
importFrom(Seqinfo,genome)
importFrom(Seqinfo,merge)
importFrom(Seqinfo,seqinfo)
importFrom(Seqinfo,seqlengths)
importFrom(Seqinfo,seqlevels)
importFrom(Seqinfo,seqnames)
importFrom(VariantAnnotation,ScanVcfParam)
importFrom(VariantAnnotation,info)
importFrom(VariantAnnotation,readVcf)
importFrom(crisprBase,"spacerLength<-")
importFrom(crisprBase,CrisprNuclease)
importFrom(crisprBase,cutSites)
importFrom(crisprBase,editingStrand)
importFrom(crisprBase,editingWeights)
importFrom(crisprBase,getCutSiteFromPamSite)
importFrom(crisprBase,getCutSiteRanges)
importFrom(crisprBase,getEditingSiteFromPamSite)
importFrom(crisprBase,getProtospacerRanges)
importFrom(crisprBase,hasSpacerGap)
importFrom(crisprBase,isCutting)
importFrom(crisprBase,isDnase)
importFrom(crisprBase,isRnase)
importFrom(crisprBase,motifs)
importFrom(crisprBase,nucleaseName)
importFrom(crisprBase,pamLength)
importFrom(crisprBase,pamSide)
importFrom(crisprBase,pams)
importFrom(crisprBase,prototypeSequence)
importFrom(crisprBase,spacerGap)
importFrom(crisprBase,spacerLength)
importFrom(crisprBase,targetType)
importFrom(crisprBase,weights)
importFrom(crisprBowtie,runCrisprBowtie)
importFrom(crisprScore,getCFDScores)
importFrom(crisprScore,getCRISPRaterScores)
importFrom(crisprScore,getCRISPRscanScores)
importFrom(crisprScore,getCasRxRFScores)
importFrom(crisprScore,getDeepHFScores)
importFrom(crisprScore,getEnPAMGBScores)
importFrom(crisprScore,getLindelScores)
importFrom(crisprScore,getMITScores)
importFrom(crisprScore,getRuleSet1Scores)
importFrom(crisprScore,getRuleSet3Scores)
importFrom(methods,as)
importFrom(methods,callNextMethod)
importFrom(methods,is)
importFrom(methods,new)
importFrom(reticulate,use_condaenv)
importFrom(rtracklayer,import.bw)
importFrom(stats,complete.cases)
importFrom(txdbmaker,makeTxDbFromEnsembl)
importFrom(txdbmaker,makeTxDbFromGFF)
importFrom(utils,adist)
importFrom(utils,data)
importFrom(utils,globalVariables)
