data(grListExample, package="crisprDesign")
data(grRepeatsExample, package="crisprDesign")
data(guideSetExample, package="crisprDesign")
data(guideSetExampleFullAnnotation, package="crisprDesign")
data(tssObjectExample, package="crisprDesign")
data("SpCas9", package="crisprBase")
